@@ -9,7 +9,7 @@ from enum import Enum
 
 from . import shellpattern
 from .helpers import clean_lines
-from .helpers import Error
+from .helpers.errors import Error
 
 
 def parse_patternfile_line(line, roots, ie_commands, fallback):
@@ -147,7 +147,7 @@ class PatternMatcher:
         in self.fallback is returned (defaults to None).
 
         """
-        path = normalize_path(path)
+        path = normalize_path(path).lstrip(os.path.sep)
         # do a fast lookup for full path matches (note: we do not count such matches):
         non_existent = object()
         value = self._path_full_patterns.get(path, non_existent)
@@ -201,7 +201,7 @@ class PatternBase:
         return matches
 
     def __repr__(self):
-        return '%s(%s)' % (type(self), self.pattern)
+        return f'{type(self)}({self.pattern})'
 
     def __str__(self):
         return self.pattern_orig
@@ -219,7 +219,7 @@ class PathFullPattern(PatternBase):
     PREFIX = "pf"
 
     def _prepare(self, pattern):
-        self.pattern = os.path.normpath(pattern)
+        self.pattern = os.path.normpath(pattern).lstrip(os.path.sep)  # sep at beginning is removed
 
     def _match(self, path):
         return path == self.pattern
@@ -239,7 +239,9 @@ class PathPrefixPattern(PatternBase):
     PREFIX = "pp"
 
     def _prepare(self, pattern):
-        self.pattern = os.path.normpath(pattern).rstrip(os.path.sep) + os.path.sep
+        sep = os.path.sep
+
+        self.pattern = (os.path.normpath(pattern).rstrip(sep) + sep).lstrip(sep)  # sep at beginning is removed
 
     def _match(self, path):
         return (path + os.path.sep).startswith(self.pattern)
@@ -257,7 +259,7 @@ class FnmatchPattern(PatternBase):
         else:
             pattern = os.path.normpath(pattern) + os.path.sep + '*'
 
-        self.pattern = pattern
+        self.pattern = pattern.lstrip(os.path.sep)  # sep at beginning is removed
 
         # fnmatch and re.match both cache compiled regular expressions.
         # Nevertheless, this is about 10 times faster.
@@ -281,7 +283,7 @@ class ShellPattern(PatternBase):
         else:
             pattern = os.path.normpath(pattern) + sep + "**" + sep + "*"
 
-        self.pattern = pattern
+        self.pattern = pattern.lstrip(sep)  # sep at beginning is removed
         self.regex = re.compile(shellpattern.translate(self.pattern))
 
     def _match(self, path):
@@ -294,7 +296,7 @@ class RegexPattern(PatternBase):
     PREFIX = "re"
 
     def _prepare(self, pattern):
-        self.pattern = pattern
+        self.pattern = pattern  # sep at beginning is NOT removed
         self.regex = re.compile(pattern)
 
     def _match(self, path):
@@ -313,7 +315,7 @@ _PATTERN_CLASSES = {
     ShellPattern,
 }
 
-_PATTERN_CLASS_BY_PREFIX = dict((i.PREFIX, i) for i in _PATTERN_CLASSES)
+_PATTERN_CLASS_BY_PREFIX = {i.PREFIX: i for i in _PATTERN_CLASSES}
 
 CmdTuple = namedtuple('CmdTuple', 'val cmd')
 
@@ -337,7 +339,7 @@ def get_pattern_class(prefix):
     try:
         return _PATTERN_CLASS_BY_PREFIX[prefix]
     except KeyError:
-        raise ValueError("Unknown pattern style: {}".format(prefix)) from None
+        raise ValueError(f"Unknown pattern style: {prefix}") from None
 
 
 def parse_pattern(pattern, fallback=FnmatchPattern, recurse_dir=True):
@@ -392,7 +394,7 @@ def parse_inclexcl_command(cmd_line_str, fallback=ShellPattern):
         try:
             val = get_pattern_class(remainder_str)
         except ValueError:
-            raise argparse.ArgumentTypeError("Invalid pattern style: {}".format(remainder_str))
+            raise argparse.ArgumentTypeError(f"Invalid pattern style: {remainder_str}")
     else:
         # determine recurse_dir based on command type
         recurse_dir = command_recurses_dir(cmd)
